from django.contrib.gis.db.models.fields import MultiLineStringField, MultiPolygonField
from django.db import models

from .osmimport_models import (  # noqa: F401
    PlanetOsmLine,
    PlanetOsmPoint,
    PlanetOsmPolygon,
)


class OsmHighway(models.Model):
    geom = MultiLineStringField(srid=3857)
    name = models.TextField(null=True, blank=True)
    highway = models.TextField()


class OsmAdminBoundary(models.Model):
    geom = MultiLineStringField(srid=3857)
    name = models.TextField(null=True, blank=True)


class OsmIslands(models.Model):
    geom = MultiLineStringField(srid=3857)
    name = models.TextField(null=True, blank=True)


class OsmIslandsAreas(models.Model):
    geom = MultiPolygonField(srid=3857)
    name = models.TextField(null=True, blank=True)


class FacebookAiRoad(models.Model):
    way_fbid = models.BigIntegerField()
    geom = MultiLineStringField(srid=3857)
    highway = models.TextField()



from django.contrib import admin

from .models import OsmAdminBoundary, OsmHighway, OsmIslands

# Register your models here.


# @admin.register(PlanetOsmLine)
# class PlanetOsmLineAdmin(admin.ModelAdmin):
#     list_display = ["name", "boundary", "highway", "waterway"]


# @admin.register(PlanetOsmPoint)
# class PlanetOsmPointAdmin(admin.ModelAdmin):
#     list_display = ["name", "place", "amenity", "natural"]


# @admin.register(PlanetOsmPolygon)
# class PlanetOsmPolygonAdmin(admin.ModelAdmin):
#     list_display = ["name", "place"]


@admin.register(OsmHighway)
class OsmHighwayAdmin(admin.ModelAdmin):
    list_display = ["name", "highway"]


@admin.register(OsmIslands)
class OsmIslandsAdmin(admin.ModelAdmin):
    list_display = ["name"]


@admin.register(OsmAdminBoundary)
class OsmAdminBoundaryAdmin(admin.ModelAdmin):
    list_display = ["name"]



from functools import reduce
from operator import add
from typing import Any, List
from django.apps import apps

from django.db import connection
from django.http import JsonResponse
from django.http.response import HttpResponse
from django.views import View
from django.views.generic.base import TemplateView
from django.contrib.gis.db.models.functions import Centroid, Transform
from dataclasses import asdict
from psycopg2 import sql

from annotations import GeoJsonSerializer, MultiGeoJsonSerializer

from .models import (
    FacebookAiRoad,
    OsmAdminBoundary,
    OsmHighway,
    OsmIslands,
    OsmIslandsAreas,
)
from .tilegenerator import MvtQuery, Tile


class ExampleMapView(TemplateView):
    template_name = "leaflet_tile_layers.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["map_view"] = f"[-8.556416, 125.5524171], 14"
        return context


class TileLayerView(View):
    """
    Handle generation and views of tile layers
    For urls.py you likely want two entries for easier reversing in templates
    >>> path("mylayer/<int:zoom>/<int:x>/<int:y>.pbf", MyTileLayerView.as_view()), name="mylayer"),
    >>> path("mylayer", MyTileLayerView.as_view(), name="mylayer"),
    """

    layers: List[MvtQuery] = []

    def get_layers(self, tile: Tile) -> List[MvtQuery]:
        """
        Override for dynamically determining tile content
        based on tile properties (most likely zoom)
        """

        def filter_for_zoom(layer: MvtQuery):
            if layer.min_render_zoom and tile.zoom < layer.min_render_zoom:
                return False
            if layer.max_render_zoom and tile.zoom > layer.max_render_zoom:
                return False

        return list(filter(filter_for_zoom, self.layers))

    def _generate_tile(self, tile: Tile) -> List[Any]:

        params = asdict(tile)

        with connection.cursor() as cursor:
            tiles = []
            for layer in self.get_layers(tile):
                query = layer.as_mvt()
                cursor.execute(query, params)
                tile_response = cursor.fetchone()
                if tile_response:
                    tiles += tile_response[0]

            return tiles

    def get(self, request, *args, **kwargs):
        tile = Tile(**kwargs)  # Expect to receive zoom, x, and y in kwargs
        tiles = reduce(add, self._generate_tile(tile), b'')
        return HttpResponse(content=tiles, content_type="application/binary")


class RoadLayerView(TileLayerView):
    def get_layers(self, tile: Tile):
        layers = []
        for road_class, min_zoom in [
            # ("steps", 5),
            # ("bus_guideway", 5),
            # ("footway", 5),
            # ("services", 5),
            ("trunk", 3),
            ("road", 12),
            ("secondary", 5),
            ("trunk_link", 3),
            ("tertiary", 10),
            ("secondary_link", 5),
            ("tertiary_link", 10),
            ("primary", 3),
            ("residential", 12),
            ("primary_link", 12),
            ("track", 12),
            ("service", 12),
            ("unclassified", 12),
            ("path", 12),
        ]:
            layers.append(
                MvtQuery(
                    table=OsmHighway._meta.db_table,
                    attributes=["name", "highway"],
                    filters=sql.SQL("{} = {}").format(sql.Identifier("highway"), sql.Literal(road_class)),
                    layer=road_class,
                    min_render_zoom=min_zoom,
                    transform=False,
                )
            )
        return layers


class AdminBoundaryLayerView(TileLayerView):
    layers = [MvtQuery(table=OsmAdminBoundary._meta.db_table, attributes=["name"], layer="admin_boundary")]


class IslandsLayerView(TileLayerView):
    layers = [MvtQuery(table=OsmIslands._meta.db_table, attributes=["name"], layer="islands")]


class IslandsAreaLayerView(TileLayerView):
    layers = [MvtQuery(table=OsmIslandsAreas._meta.db_table, attributes=["name"], layer="islands")]


class FacebookAiLayerView(TileLayerView):
    layers = [MvtQuery(table=FacebookAiRoad._meta.db_table, attributes=["highway"], layer="facebookai")]


class BuildingPolygon(TileLayerView):
    layers = [MvtQuery(table="osmflex_buildingpolygon", attributes=["name", "osm_id", "osm_subtype"], layer="buildings", pk="osm_id")]


class MajorRoads(TileLayerView):
    layers = [MvtQuery(table="osmflex_roadmajor", attributes=["name", "osm_type"], layer="roads", pk="osm_id")]


class Hospitals(View):
    def get(self, request, *args, **kwargs):
        AmenityPoint = apps.get_model("osmflex", "AmenityPoint")
        AmenityPolygon = apps.get_model("osmflex", "AmenityPolygon")

        points = AmenityPoint.objects.filter(osm_type__in=["hospital", "clinic"]).annotate(geom_for_json=Transform("geom", 4326))
        centroids = AmenityPolygon.objects.filter(osm_type__in=["hospital", "clinic"]).annotate(geom_centroid=Transform(Centroid("geom"), 4326))

        response = MultiGeoJsonSerializer(
            [
                GeoJsonSerializer(points, "geom_for_json", ("name", "osm_type", "address")),
                GeoJsonSerializer(centroids, "geom_centroid", ("name", "osm_type", "address")),
            ]
        )

        return JsonResponse(data=asdict(response.to_collection()), content_type="application/json")


class Aeroways(View):
    def get(self, request, *args, **kwargs):
        point = apps.get_model("osmflex", "InfrastructurePoint")
        polygon = apps.get_model("osmflex", "InfrastructurePolygon")

        points = point.objects.filter(osm_type__in=["aeroway"]).annotate(geom_for_json=Transform("geom", 4326))
        centroids = polygon.objects.filter(osm_type__in=["aeroway"]).annotate(geom_centroid=Transform(Centroid("geom"), 4326))

        response = MultiGeoJsonSerializer(
            [
                GeoJsonSerializer(points, "geom_for_json", ("name", "osm_type")),
                GeoJsonSerializer(centroids, "geom_centroid", ("name", "osm_type")),
            ]
        )

        return JsonResponse(data=asdict(response.to_collection()), content_type="application/json")



from dataclasses import dataclass, field
from itertools import chain
from typing import List, Optional, Sequence

PgFunction = str
from psycopg2 import sql
from psycopg2.sql import Literal as L, Identifier as I

from operator import add
from functools import reduce

@dataclass
class Tile:
    """
    Simple dataclass representing a "tile" instance and additional metadata
    for tile generation: buffer, extent and margin
    http://postgis.net/docs/ST_TileEnvelope.html

    """

    zoom: int
    x: int
    y: int
    buffer: int = 64
    extent: int = 4096

    @staticmethod
    def tile_envelope() -> sql.Composable:
        return sql.SQL("ST_TileEnvelope(%(zoom)s, %(x)s, %(y)s)")

    @staticmethod
    def tile_envelope_margin() -> sql.Composable:
        return sql.SQL("ST_TileEnvelope(%(zoom)s, %(x)s, %(y)s, margin => (%(buffer)s / %(extent)s))")


@dataclass
class MvtQuery:
    """
    This is a SQL query generator based on the example
    at https://postgis.net/docs/manual-dev/ST_AsMVT.html
    """

    table: str
    attributes: List[str] = field(default_factory=list)
    filters: Sequence[sql.Composable] = field(default_factory=list)
    transform: bool = False  # Set to True if source srid is not 3857, but beware performanc
    field: str = "geom"
    pk: str = "id"
    layer: str = "default"
    min_render_zoom: Optional[int] = None
    max_render_zoom: Optional[int] = None

    @property
    def json_attributes(self) -> sql.Composed:
        """
        Create a "JSONB_BUILD_OBJECT" clause
        """
        if not self.attributes:
            return sql.Composed(sql.SQL(""))
        params = None
        for a in self.attributes:
            if not params:
                params = sql.Literal(a)
            else:
                params += sql.Literal(a)  # Name of the key is the same as the field
            params += sql.Identifier(a)  # The field to use as the value for the JSON
        composed_params = params.join(", ")  # type: ignore
        return sql.SQL("jsonb_build_object({})").format(composed_params)

    @property
    def transformed_geom(self) -> sql.Composable:
        """
        Return an `ST_TRANSFORM` clause
        https://postgis.net/docs/ST_Transform.html
        """
        if self.transform:
            return sql.SQL("ST_TRANSFORM({field}, 3857)").format(sql.Identifier(self.field))
        return sql.Identifier(self.field)

    @property
    def alias(self) -> sql.Composable:
        """
        The name of the query alias to use for postgresql
        """
        return sql.Identifier(f"mvt_{self.layer}")

    @property
    def where(self) -> sql.Composable:
        if self.filters:
            where_clause = sql.SQL(" AND ") + sql.SQL(" AND ").join([self.filters])
        else:
            where_clause = sql.SQL("")
        return where_clause

    @property
    def as_mvtgeom(self) -> sql.Composable:

        if self.filters:
            where_clause = sql.SQL(" AND ") + sql.SQL(" AND ").join([self.filters])
        else:
            where_clause = sql.SQL("")

        return sql.SQL(
            """
            ST_AsMVTGeom(
                {g},
                {e},
                extent => %(extent)s,
                buffer => %(buffer)s
            ) AS geom, {pk}, {json_attributes}
            FROM {t}
            WHERE {g} && {m} {where}
            """
        ).format(
            g=self.transformed_geom,
            m=Tile.tile_envelope_margin(),
            e=Tile.tile_envelope(),
            t=sql.Identifier(self.table),
            # Properties of "self"
            pk=sql.Identifier(self.pk),
            json_attributes=self.json_attributes,
            where=self.where,
        )

    def as_mvt(self) -> PgFunction:
        outer_query = sql.SQL("""WITH {alias} AS (SELECT {inner_query} ) SELECT ST_AsMVT( {alias}.*, {layer}, %(extent)s, 'geom', {pk}) FROM {alias}""")
        parameters = dict(alias=self.alias, layer=sql.Literal(self.layer), inner_query=self.as_mvtgeom, pk=sql.Literal(self.pk))
        return outer_query.format(**parameters)
